/****************************************************************************
 * xpf_interface.c:The interface provided to the user of xpf_dags.c.	    *
 *		All functions are "callable": there are no upcalls (Larry's *
 *		suggestion). Put in more code from xpf_protl.c in here.	    *
 ****************************************************************************/

#include "xpf_internal.h"

#define DOPRINT(s)
#define NOPRINT(s)
#define NOWPRINT(s)
#define ERRPRINT(s)

extern int	strnmaskordercmp();
extern int	bmaskordercmp();
extern int	bordercmp();
extern int	memmaskcpy();
extern int	xpfGetPositionFromPatternList();
extern int	xpfLaunchProcesses();
extern int	xpfDeletePatternList();
extern void	xpfInsertPATNode();
extern PATNode	*xpfInsertPatternList();
extern HashEntry *xpfGetHashEntry();
extern void	xpfDestroyLivePATTree();
void		xpfPrintPATTree();
extern int	xpfOrderPatternLists();
extern void	xpfTerminateProcesses();

/*
 * Must be called first. Returns the "state" which should be passed as
 * a parameter to all other interface calls. Two operations on one state
 * cannot be done in parallel (implicit serialization). However,
 * operations on two XPFs can be done in parallel: hence the device
 * driver can use one XPFObject for each network-interface.
 */

int
xpfCreateObject(state)
	XPFObject	*state;
{
	if ((*state = (XPFObject)xpfMalloc(sizeof(XPFState))) == NULL) return FALSE;
	xpfBzero((char *)*state, sizeof(XPFState));
	((XPFState *)*state)->xpfMaxAssociations = XPF_DEFAULT_MAXASSOCIATIONS;
	((XPFState *)*state)->xpfGranularity = XPF_DEFAULT_GRANULARITY;
	((XPFState *)*state)->xpfTimer = XPF_DEFAULT_TIMER;
	return TRUE;
}

/*
 * Completely ruins the "state". The parameter is invalid after return.
 * Returns the set of deletedmessages and deletedinfos in specified pointers.
 */

void
xpfDestroyObject(state
#if	MERGEINFO
		, pdeletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
		, pdeletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
		)
	XPFObject	state;
#if	MERGEINFO
	XPFInfoList	**pdeletedinfo;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	**pdeletedmessagelist;
#endif	DELAYEDMSGS && STOREMSGS
{
	while(((XPFState *)state)->xpfAssociationSet != NULL)
		xpfDestroyLivePATTree(&((XPFState *)state)->xpfAssociationSet
#if	MERGEINFO
					, pdeletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					, pdeletedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					);
	xpfFree((char *)state, sizeof(XPFObject));
}

/*
 * Sets up an association between patternlist and info so that
 * messages matching the patternlist can be mapped to the info.
 * Returns TRUE if successful, FALSE otherwise.
 */

int
xpfSetupAssociation(state, patternlist, info)
	XPFObject	state;
	XPFPattern	*patternlist;
#if	MERGEINFO
	XPFInfoList	*info;
#else	MERGEINFO
	XPFInformation	info;
#endif	MERGEINFO
{
	PATNode		*retval;
#if	MERGEINFO
	XPFInfoList	**destinfo, *srcinfo;
#endif	MERGEINFO

	DOPRINT(xpfPrintf("Before Setup:\n");)
	DOPRINT(xpfPrintPATTree(((XPFState *)state)->xpfAssociationSet, 0);)
	if (((XPFState *)state)->xpfNumAssociations >= ((XPFState *)state)->xpfMaxAssociations) return FALSE;
	if (((XPFState *)state)->xpfAssociationSet == NULL)
		retval = xpfInsertPatternList(patternlist,
						patternlist->filterlist,
						&((XPFState *)state)->xpfAssociationSet,
						NULL,
						((XPFState *)state)->xpfGranularity,
						XPF_DAGS_SRCNEWDESTNEW,
						XPF_FLAG_USER);
	else retval = xpfInsertPatternList(patternlist,
					patternlist->filterlist,
					&((XPFState *)state)->xpfAssociationSet,
					&((XPFState *)state)->xpfAssociationSet->common.pattern.filterlist,
					((XPFState *)state)->xpfGranularity,
					XPF_DAGS_SRCNEWDESTNEW,
					XPF_FLAG_USER);
	if (retval == NULL) return FALSE;

#if	MERGEINFO
	destinfo = &retval->info;
	for (srcinfo = info; srcinfo != XPF_INFO_NULL; srcinfo = srcinfo->next, destinfo = &(*destinfo)->next) {
		*destinfo = (XPFInfoList *)xpfMalloc(sizeof(XPFInfoList));
		(*destinfo)->information = srcinfo->information;
	}
	*destinfo = XPF_INFO_NULL;
#else	MERGEINFO
	retval->info = info;
#endif	MERGEINFO
	((XPFState *)state)->xpfNumAssociations ++;
	DOPRINT(xpfPrintf("info = %x numassoc = %x\n", info, ((XPFState *)state)->xpfNumAssociations);)
	NOWPRINT(xpfPrintf("After Setup:\n");)
	NOWPRINT(xpfPrintPATTree(((XPFState *)state)->xpfAssociationSet, 0);)
	return TRUE;
}

/*
 * Sets up an association between prefixpat.newpat and info
 * such that its suffix is the same as that of prefixpat.findsuffixpat.
 * This allows the user to manipulate patterns so that a logical OR
 * can be realized. Returns TRUE if successful, FALSE otherwise.
 *
 * Should I use more meaningful return values (find prefix of new,
 * suffixes not identical, some info/msgs got deleted!). Uses the
 * fact that a HASHENTRY cannot directly point to another HASHENTRY and
 * HASHENTRYs have only one reference (unless initial): so the ->and
 * of a PATNode returned by Get/Insert will NEVER be a HASHENTRY.
 */

int
xpfSetupAssociationSuffix(state, prefixpat, newpat, findsuffixpat, info)
	XPFObject	state;
	XPFPattern	*prefixpat;
	XPFPattern	*newpat;
	XPFPattern	*findsuffixpat;
#if	MERGEINFO
	XPFInfoList	*info;
#else	MERGEINFO
	XPFInformation	info;
#endif	MERGEINFO
{
	PATNode		*retval;
	PATNode		**proot = &(((XPFState *)state)->xpfAssociationSet);
	PATNode		**iproot = NULL;
	XPFFilter	**idestfilter = NULL;
	int		ilocation;
	PATNode		**jproot = NULL, *suffixnode = NULL;
	XPFFilter	**jdestfilter = NULL;
#if	MERGEINFO
	XPFInfoList	**destinfo, *srcinfo;
#endif	MERGEINFO
	int		ret;
#if	MERGEINFO
	XPFInfoList	*deletedinfo;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	*deletedmessages;
#endif	DELAYEDMSGS && STOREMSGS

	DOPRINT(xpfPrintf("In SetupSuffix\n");)
	if (((XPFState *)state)->xpfNumAssociations >= ((XPFState *)state)->xpfMaxAssociations) return FALSE;
	if (((XPFState *)state)->xpfAssociationSet == NULL) {	/* degenerates into SetupAssociation, findsuffix is useless */
		if (prefixpat == NULL) prefixpat = newpat;
		else {
			XPFPattern	**p = &prefixpat->next;

			while(*p != NULL) p = &(*p)->next;
			*p = newpat;	/* append */
			newpat->header.headertype |= XPF_HEADER_CURRENT;
		}

		return xpfSetupAssociation(state, prefixpat, info);
	}
	if ((prefixpat == NULL) && (findsuffixpat == NULL)) return xpfSetupAssociation(state, newpat, info);
	if (xpfOrderPatternLists(findsuffixpat, newpat, ((XPFState *)state)->xpfGranularity) == XPF_DAGS_SRCPREFIX) return FALSE;	/* cycle */

	if (prefixpat == NULL) {
		iproot = &(((XPFState *)state)->xpfAssociationSet);
		ilocation = XPF_DAGS_SRCNEWDESTNEW;
		if (*iproot == NULL) {
			idestfilter = NULL;
		}
		else {
			idestfilter = &(*iproot)->common.pattern.filterlist;
		}
	}
	else {
		/* No need to store headertype for prefixpat since it is not used again in this function */
		if (*proot == NULL)
			ret = xpfGetPositionFromPatternList(state,
							prefixpat,
							prefixpat->filterlist,
							proot,
							NULL,
							((XPFState *)state)->xpfGranularity,
							XPF_DAGS_SRCNEWDESTNEW,
							&iproot,
							&idestfilter);
		else 
			ret = xpfGetPositionFromPatternList(state,
							prefixpat,
							prefixpat->filterlist,
							proot,
							&(*proot)->common.pattern.filterlist,
							((XPFState *)state)->xpfGranularity,
							XPF_DAGS_SRCNEWDESTNEW,
							&iproot,
							&idestfilter);
		xpfRestoreHeaders(((XPFState *)state));	/* maximum without message argument: we don't need to look at matched nodes */
		if (ret == FALSE) return FALSE;
		if (iproot == NULL) return FALSE;

		/* I did match something so new and suffix have current hdrs */
		newpat->header.headertype |= XPF_HEADER_CURRENT;
		findsuffixpat->header.headertype |= XPF_HEADER_CURRENT;

		if (*iproot != NULL) {
			if ((*iproot)->nodetype & XPF_DAGS_HASHENTRY) {	/* skip over the matched hash entry */
				iproot = &(*iproot)->and;
				ilocation = XPF_DAGS_SRCNEWDESTNEW;
				if (*iproot != NULL) {
					idestfilter = &(*iproot)->common.pattern.filterlist;
				}
				else {
					idestfilter = NULL;
				}
			}
			else {	/* go one step ahead for insert-after */
				if (idestfilter == NULL) return FALSE;	/* some error! */
				if ((*idestfilter)->next == NULL) {
					ilocation = XPF_DAGS_SRCNEWDESTNEW;
					if ((*iproot)->and == NULL) {
						iproot = &(*iproot)->and;
						idestfilter = NULL;
					}
					else {
						iproot = &(*iproot)->and;
						idestfilter = &(*iproot)->common.pattern.filterlist;
					}
				}
				else {
					ilocation = XPF_DAGS_SRCNEWDESTOLD;
					idestfilter = &(*idestfilter)->next;
				}
			}
		}
		else {
			ilocation = XPF_DAGS_SRCNEWDESTNEW;
			idestfilter = NULL;
		}
	}
	NOWPRINT(xpfPrintf("prefix=%x\n", ilocation);)
	NOWPRINT(xpfPrintPATTree((*iproot), 0);)

	/*
	 * iproot and idestfiliter will remain the same since the above insertion is done AFTER that point: their *s might change though!
	 * The code below is very similar to above but a bit different since though the prefix can't be NULL, the suffix can be NULL (as if xpfSetup).
	 * No need to store headertype for findsuffixpat since it is not used again in this function.
	 */
	if ((ret = xpfGetPositionFromPatternList(state,
					findsuffixpat,
					findsuffixpat->filterlist,
					iproot,
					idestfilter,
					((XPFState *)state)->xpfGranularity,
					ilocation /* XPF_DAGS_SRCNEWDESTNEW */,
					&jproot,
					&jdestfilter)) == FALSE) {
		ERRPRINT(xpfPrintf("xpfSetupAssociationSuffix: warning! suffix pattern not found.\n");)
		jproot = NULL;
		jdestfilter = NULL;
	}
	else {
		if (*jproot != NULL) {
			if ((*jproot)->nodetype & XPF_DAGS_HASHENTRY) {	/* skip over the matched hash entry */
				jproot = &(*jproot)->and;
				if (*jproot != NULL) {
					jdestfilter = &(*jproot)->common.pattern.filterlist;
				}
				else {
					jdestfilter = NULL;
				}
			}
			else {	/* go one step ahead for insert-after */
				if (jdestfilter == NULL) return FALSE;	/* some error! */
				if ((*jdestfilter)->next == NULL) {
					if ((*jproot)->and == NULL) {
						jproot = &(*jproot)->and;
						jdestfilter = NULL;
					}
					else {
						jproot = &(*jproot)->and;
						jdestfilter = &(*jproot)->common.pattern.filterlist;
					}
				}
				else jdestfilter = &(*jdestfilter)->next;
			}
			suffixnode = *jproot;
		}
		else {
			jdestfilter = NULL;
			suffixnode = NULL;
		}
	}
	xpfRestoreHeaders(((XPFState *)state));	/* maximum without message argument: we don't need to look at matched nodes */
	NOWPRINT(xpfPrintf("findsuffix=\n");)
	NOWPRINT(if (suffixnode != NULL) xpfPrintPATTree(suffixnode, 0);)

	if ((retval = xpfInsertPatternList(newpat,
					newpat->filterlist,
					iproot,
					idestfilter,
					((XPFState *)state)->xpfGranularity,
					ilocation /* XPF_DAGS_SRCNEWDESTNEW */,
					XPF_FLAG_USER)) == NULL) return FALSE;
	NOWPRINT(xpfPrintf("after insert=\n");)
	NOWPRINT(xpfPrintPATTree(((XPFState *)state)->xpfAssociationSet, 0);)

	/* Now set the information before setting the suffix */
#if	MERGEINFO
	destinfo = &retval->info;
	for (srcinfo = info; srcinfo != XPF_INFO_NULL; srcinfo = srcinfo->next, destinfo = &(*destinfo)->next) {
		*destinfo = (XPFInfoList *)xpfMalloc(sizeof(XPFInfoList));
		(*destinfo)->information = srcinfo->information;
	}
	*destinfo = XPF_INFO_NULL;
#else	MERGEINFO
	retval->info = info;
#endif	MERGEINFO
	((XPFState *)state)->xpfNumAssociations ++;

	if (retval->and != NULL) ERRPRINT(xpfPrintf("xpfSetupAssociationSuffix: warning! suffixes are not identical.\n"));
#if	MERGEINFO
	deletedinfo = XPF_INFO_NULL;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	deletedmessages = NULL;
#endif	DELAYEDMSGS && STOREMSGS

	if (suffixnode != NULL) {	/* difference from launch programs: it is not enough to just look at suffixnode as an insertion point */
		if ( (jdestfilter == NULL) || (*jdestfilter == NULL) ||	/* last 2 can't happen! */
		     ((*jproot)->common.pattern.filterlist == *jdestfilter) )
			xpfInsertPATNode(&retval->and, suffixnode, ((XPFState *)state)->xpfGranularity
#if	MERGEINFO
					, &deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					, &deletedmessages
#endif	DELAYEDMSGS && STOREMSGS
					);
			/* we stopped at the end of a hash-entry or at the beginning of a (normal) node */
		else {	/* we stopped at the middle of a (normal) node */
			PATNode	*n;

			ERRPRINT(xpfAssert((*jdestfilter) != NULL));
			xpfCreatePATNode(n);
			n->common.pattern.flags = (*jproot)->common.pattern.flags;
			(*jproot)->common.pattern.flags &= XPF_FLAG_RESIDUE;
#if	VARHEADER
			n->common.pattern.header.size.variable = NULL;
#endif	VARHEADER
			COPY_HEADER(&n->common.pattern.header, &(*jproot)->common.pattern.header);
			n->common.pattern.header.headertype |= XPF_HEADER_CURRENT;
			n->common.pattern.filterlist = (*jdestfilter);
			(*jdestfilter) = NULL;
			n->common.pattern.programlist = (*jproot)->common.pattern.programlist;
			(*jproot)->common.pattern.programlist = NULL;

			n->and = suffixnode->and;
			suffixnode->and = n;
			n->refcount ++;	/* refcount of suffixnode->and is unchanged */
			n->or = suffixnode->or;
			if (n->or != NULL) n->or->refcount ++;

			xpfInsertPATNode(&retval->and, n, ((XPFState *)state)->xpfGranularity
#if	MERGEINFO
					, &deletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					, &deletedmessages
#endif	DELAYEDMSGS && STOREMSGS
					);
		}
	}
	/* else there is nothing to do: suffix remains NULL */

#if	MERGEINFO
	if (deletedinfo) ERRPRINT(xpfPrintf("xpfSetupAssociationSuffix: warning! some infos got deleted\n"));
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	if (deletedmessages) ERRPRINT(xpfPrintf("xpfSetupAssociationSuffix: warining! some messages got deleted.\n"));
#endif	DELAYEDMSGS && STOREMSGS
	NOWPRINT(xpfPrintf("After set suffix=\n");)
	NOWPRINT(xpfPrintPATTree(((XPFState *)state)->xpfAssociationSet, 0);)

	return TRUE;
}

/*
 * Removes the association corr. to patternlist, sets *pdeletedinfo to
 * the info(s) that are no-longer associated with any pattern and
 * sets *pdelayedmessagelist to the corr. list of delayed messages. Also
 * terminates programs (if any) generated by this pattern.
 *
 * Returns TRUE if the whole patternlist got deleted successfully;
 * FALSE o/wise -- pdeletedinfo and pdelayedmessagelist might contain deleted
 * stuff EVEN if Remove was not entirely successful -- e.g., this
 * can happen if only a part of the patternlist->filterlist matched.
 *
 * NOTE: If the patternlist is "A and B", and "B" gets deleted, then
 * the info and messagelist associated with this pattern are
 * NOT returned. They are returned only when "A" also gets deleted
 * and the info and messagelist aren't associated w/ anything.
 * It is upto the user to figure out what patterns these correspond to
 * and destroy (free) them, re-establish new associations (setup),
 * or re-process the set of messages (lookup) through the filter.
 */

int
xpfRemoveAssociation(state, patternlist
#if	MERGEINFO
			, pdeletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
			, pdelayedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
			)
	XPFObject	state;
	XPFPattern	*patternlist;
#if	MERGEINFO
	XPFInfoList	**pdeletedinfo;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	**pdelayedmessagelist;
#endif	DELAYEDMSGS && STOREMSGS
{
	PATNode		**proot;
	PATNode		**iproot = NULL;
	XPFFilter	**idestfilter = NULL;
	PATNode		***matchednodes;	/* to simplify coding */
	int		matchednum;
	int		stored_headertype[XPF_DAGS_MAXDEPTH];
	int		num_pat=0, i, k;
	XPFPattern	*temppat;

	if (((XPFState *)state)->xpfAssociationSet != NULL) {
		for (temppat=patternlist; temppat!=NULL; temppat=temppat->next)
			stored_headertype[num_pat++] = temppat->header.headertype;

		/* Use this function to just get the set of matched nodes: patternlist is used again in Delete below, so store/restore pattern's headertype */
		xpfGetPositionFromPatternList((XPFState *)state,
						patternlist,
						patternlist->filterlist,
						&((XPFState *)state)->xpfAssociationSet,
						&((XPFState *)state)->xpfAssociationSet->common.pattern.filterlist,
						((XPFState *)state)->xpfGranularity,
						XPF_DAGS_SRCNEWDESTNEW,
						&iproot,
						&idestfilter);
		
		for (temppat=patternlist,i=0; temppat!=NULL,i<num_pat; temppat=temppat->next,i++)
			temppat->header.headertype = stored_headertype[i];

		matchednodes = (((XPFState *)state)->xpfMatchedNodes);
		matchednum = (((XPFState *)state)->xpfMatchedNum);
		if ( (matchednum <= 0) || (iproot == NULL) || (*iproot == NULL) || 
		     (((*iproot)->nodetype != XPF_DAGS_HASHENTRY) &&
		      ((idestfilter == NULL) || (*idestfilter == NULL))
		     )
		   ){
			xpfRestoreHeaders(((XPFState *)state));
			return FALSE;
		}

		DOPRINT(xpfPrintPATTree(*iproot, 0);)

		/* Logic is IDENTICAL to that of Launch() processes's proot calculation but we have to deal with hash-nodes: 30/9/94 */
		if ((*matchednodes[matchednum - 1])->nodetype & XPF_DAGS_HASHENTRY) {
			proot = ((matchednum - 2 > 0) ? &(*matchednodes[matchednum - 3])->and : &((XPFState *)state)->xpfAssociationSet) ;
		}
		else proot = ( (matchednum - 1 > 0) ? &(*matchednodes[matchednum - 2])->and : &((XPFState *)state)->xpfAssociationSet) ;
		if ((*iproot)->common.pattern.header.headertype & XPF_HEADER_CURRENT) {
			if (matchednum - 1 > 0) {
				/*
				 * It is guaranteed that these two calculations will never yield a hashentry (since process/current flags are
				 * same in both hashentries and their parent nodes, so that we never quit from the while loop in the middle).
				 */
				/* must to go new proot if prev node was a process since I must insert at the first node among the processes */
				if ((*matchednodes[matchednum - 2])->nodetype & XPF_DAGS_PROCESS) {
					k = matchednum - 3;
					while ((k>0) && ((*matchednodes[k])->nodetype & XPF_DAGS_PROCESS)) k--;
					/* insert into the first matched process node */
					if (k > 0) proot = &(*matchednodes[k])->and;
					else proot = &((XPFState *)state)->xpfAssociationSet;
				}
				/* continue further back until I come to a node that is not current and insert programs right after it (into its and) */
				else {
					k = matchednum - 2;
					while ((k>0) && ((*matchednodes[k])->common.pattern.header.headertype & XPF_HEADER_CURRENT)) k--;
					/* insert into the first matched non-current header node */
					if (k > 0) proot = &(*((XPFState *)state)->xpfMatchedNodes[k])->and;
					else proot = &((XPFState *)state)->xpfAssociationSet;
				}
			}
			else proot = &((XPFState *)state)->xpfAssociationSet;	/* nowhere else to go */
		}

		ERRPRINT(xpfAssert(*proot != NULL));
		ERRPRINT(xpfAssert(!((*proot)->nodetype & XPF_DAGS_HASHENTRY)));

		xpfTerminateProcesses(proot,
					-1,
					((*iproot)->nodetype & XPF_DAGS_HASHENTRY) ? (*iproot)->common.hash.alarm : (*iproot)->common.pattern.alarm,
					((XPFState *)state)->xpfGranularity
#if	MERGEINFO
					,pdeletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					,pdelayedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					);
		DOPRINT(xpfPrintf("After Terminate:\n");)
		DOPRINT(xpfPrintPATTree(((XPFState *)state)->xpfAssociationSet, 0);)
		xpfRestoreHeaders(((XPFState *)state));

		if (xpfDeletePatternList(patternlist,
					patternlist->filterlist,
					&((XPFState *)state)->xpfAssociationSet,
					&((XPFState *)state)->xpfAssociationSet->common.pattern.filterlist,
					((XPFState *)state)->xpfGranularity,
					XPF_DAGS_SRCNEWDESTNEW
#if	MERGEINFO
					,pdeletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
					,pdelayedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
					)) {
			DOPRINT(xpfPrintf("After Good Remove:\n");)
			DOPRINT(xpfPrintPATTree(((XPFState *)state)->xpfAssociationSet, 0);)

			/*
			 * This can become -ve if the user starts removing prefixes of the original pattern instead of the whole thing at once: Or do we count
			 * as associations only those that have non-XPF_INFO_NULL mapped to them? So deletion of a prefex keeps the #of associations constant?
			 */
			((XPFState *)state)->xpfNumAssociations --;
			if (((XPFState *)state)->xpfNumAssociations < 0) ((XPFState *)state)->xpfNumAssociations = 0;
			return TRUE;
		}
		DOPRINT(xpfPrintf("After Bad Remove:\n");)
		DOPRINT(xpfPrintPATTree(((XPFState *)state)->xpfAssociationSet, 0);)
	}
	return FALSE;
}

#if	DELAYEDMSGS && STOREMSGS
/*
 * Removes the set of delayed messages at ALL patternlists that are
 * prefixes of the given "patternlist" and returns them in pdelayedmessagelist.
 * These are exactly those messages that could have been delayed due to
 * patterns in the given patternlist. The association between the patternlist
 * and its info still exists. Sets pdelayedmessagelist and returns TRUE on
 * success; Returns FALSE with pdelayedmessagelist untouched when no prefix
 * of patternlist has an association (failure).
 *
 * NOTE: The delayed messages returned in this case are EXACTLY those
 * that would have been returned by xpfRemoveAssociation, the only
 * difference being that the association is NOT removed.
 */

int
xpfRemoveDelayedMessages(state, patternlist, pdelayedmessagelist)
	XPFObject	state;
	XPFPattern	*patternlist;
	XPFMessageList	**pdelayedmessagelist;
{
	int	i;
	PATNode		**iproot = NULL;
	XPFFilter	**idestfilter = NULL;

	if (((XPFState *)state)->xpfAssociationSet != NULL) {
		XPFMessageList	**pm;

		/* Use this function to just get the set of matched nodes: no need to store headertype of patternlist since used only once */
		xpfGetPositionFromPatternList(state,
						patternlist,
						patternlist->filterlist,
						&((XPFState *)state)->xpfAssociationSet,
						&((XPFState *)state)->xpfAssociationSet->common.pattern.filterlist,
						((XPFState *)state)->xpfGranularity,
						XPF_DAGS_SRCNEWDESTNEW,
						&iproot,
						&idestfilter);

		if (((XPFState *)state)->xpfMatchedNum <= 0) {
			xpfRestoreHeaders(((XPFState *)state));
			return FALSE;
		}
		*pdelayedmessagelist = NULL;
		for (i=0; i<((XPFState *)state)->xpfMatchedNum; i++) {
			pm = &(*((XPFState *)state)->xpfMatchedNodes[i])->messagelist;
			while (*pm != NULL) pm = &(*pm)->next;
			*pm = *pdelayedmessagelist;
			*pdelayedmessagelist = (*((XPFState *)state)->xpfMatchedNodes[i])->messagelist;
			(*((XPFState *)state)->xpfMatchedNodes[i])->messagelist = NULL;
		}
		xpfRestoreHeaders(((XPFState *)state));
		return TRUE;
	}
	return FALSE;
}
#endif	DELAYEDMSGS && STOREMSGS

/*
 * Returns XPF_USER_FAILURE, XPF_USER_DELAYED, or a combination of:
 * XPF_USER_MATCHEDNEW | XPF_USER_LAUNCHED | XPF_USER_TERMINATED.
 *
 * The pointer pinfo is set to the information list associated with
 * the longest pattern that the message matches. However, if the message
 * is "delayed" for future processing (i.e., the return value is
 * XPF_USER_DELAYED), OR, there is no associated information for the
 * message or there is a serious system error (i.e., the return value is
 * XPF_USER_FAILURE), *pinfo is XPF_INFO_NULL.
 *
 * Whenever a process is triggered, it gets a higher priority than the
 * original set of patterns specified by the user for all future lookups.
 * Moreover, whenever a pattern triggers a process, the info associated
 * with that pattern is COPIED and THEN associated with the process (and
 * returned in *pinfo) this gives the user a facility to CHANGE the
 * information (s) in the info associated with a process without effecting
 * the information mapped to the (parent) pattern that generated it.
 *
 * If the message matches a pattern that has no associated programs then
 * return value & XPF_USER_MATCHEDNEW will be == 0. If it matches a process
 * generated just now by this pattern, the above & will be 1. If some processes
 * were Launched due to this pattern (including prefixes of this pattern),
 * then return value & XPF_USER_LAUNCHED will be 1. That is, if MATCHEDNEW
 * is set, then LAUNCHED has to be set, but not vice versa since there
 * might be a program for an IP fragmentation-filter, say, but not for TCP
 * and the first packet might match the TCP filter (longest match).
 *
 * The information mapped to previous processes if any is returned in the
 * pointer pdeletedinfo: if new processes were NOT generated, this
 * pointer is guaranteed to be NULL. Otherwise, it depends on the alarm values
 * specified by the user and whether any prcocesses were actually terminated.
 *
 * *pdelayedmessagelist is set to a list of delayed messages encountered
 * while trying to match the message against patterns (the messages which can
 * "hope" * to match the newly generated processes). If no such messages are
 * encountered, *pdelayedmessagelist is NULL. If the message does not spawn
 * processes (i.e., return value & XPF_USER_LAUNCHED == 0), *pdelayedmessagelist
 * is always NULL; if it spawns a fresh set of processes (the above & == 1)
 * then *pdelayedmessagelist may or may not be NULL.
 */

int
xpfLookupAssociation(state, message, length, pinfo
#if	MERGEINFO
			, pdeletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREINFO
			, pdelayedmessagelist
#endif	DELAYEDMSGS && STOREINFO
			)
	XPFObject	state;
	XPFMessage	message;
	int		length;
#if	MERGEINFO
	XPFInfoList	**pinfo;
#else	MERGEINFO
	XPFInformation	*pinfo;
#endif	MERGEINFO

#if	MERGEINFO
	XPFInfoList	**pdeletedinfo;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	**pdelayedmessagelist;
#endif	DELAYEDMSGS && STOREMSGS
{
	PATNode		*n;
	int		ret = 0;	/* default: no state change */

	DOPRINT(xpfPrintf("\nIn lookup\n");)
	DOPRINT(xpfPrintPATTree(((XPFState *)state)->xpfAssociationSet, 0);)
	((XPFState *)state)->xpfErrno = 0;

#if	MERGEINFO
	*pdeletedinfo = XPF_INFO_NULL;
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
	*pdelayedmessagelist = NULL;
#endif	DELAYEDMSGS && STOREMSGS
	xpfGetPATNodeFromMessage(((XPFState *)state), message, length, &((XPFState *)state)->xpfAssociationSet, n);
	if (n == NULL) {
		xpfRestoreAll(((XPFState *)state), message, length);
		return XPF_USER_FAILURE;
	}
	else {
		DOPRINT(xpfPrintPATTree(n, 0);)
#if	DELAYEDMSGS
		if (state->xpfErrno == XPF_DAGS_DELAY) {
			return XPF_USER_DELAYED;
		}
#endif	DELAYEDMSGS
		if (n->nodetype & XPF_DAGS_PROCESS) *pinfo = n->info;	/* ret remains 0, n must be a process */
		else {
			if ((ret = xpfLaunchProcesses(((XPFState *)state),
							pinfo
#if	MERGEINFO
							, pdeletedinfo
#endif	MERGEINFO
#if	DELAYEDMSGS && STOREMSGS
							, pdelayedmessagelist
#endif	DELAYEDMSGS && STOREMSGS
							)) == XPF_USER_FAILURE) {
				ERRPRINT(xpfPrintf("Unable to launch any process or find associated info\n"));
				xpfRestoreAll(((XPFState *)state), message, length);
				return XPF_USER_FAILURE;
			}
#if	0
			if (ret & XPF_USER_LAUNCHED) {
				xpfPrintf("\nAfter Launch\n");
				xpfPrintPATTree(((XPFState *)state)->xpfAssociationSet, 0);
			}
#endif	0
		}
	}

	DOPRINT(xpfPrintf("hdrnum=%x totsize=%x matchnum=%x \n",
			((XPFState *)state)->xpfHeaderNum,
			((XPFState *)state)->xpfHeaderTotalSize,
			((XPFState *)state)->xpfMatchedNum);)
	xpfRestoreAll(((XPFState *)state), message, length);
	return ret;
}

/*
 * Sets the flags and alarm at the matched pattern. There is no separate
 * function to set them independently, although one can write it if reqd.
 * Returns TRUE if it is able to set them properly, FALSE otherwise.
 */

int
xpfSetOptions(state, patternlist, flags, alarm)
	XPFObject	state;
	XPFPattern	*patternlist;
	int		flags;
	short		alarm;
{
	PATNode		**iproot = NULL;
	XPFFilter	**idestfilter = NULL;

	if (((XPFState *)state)->xpfAssociationSet != NULL) {
		/* Use this function to just get the set of matched nodes: no need to headertype flags of patternlist since used only once */
		xpfGetPositionFromPatternList(state,
						patternlist,
						patternlist->filterlist,
						&((XPFState *)state)->xpfAssociationSet,
						&((XPFState *)state)->xpfAssociationSet->common.pattern.filterlist,
						((XPFState *)state)->xpfGranularity,
						XPF_DAGS_SRCNEWDESTNEW,
						&iproot,
						&idestfilter);

		if (((XPFState *)state)->xpfMatchedNum <= 0) {
			xpfRestoreHeaders(((XPFState *)state));
			return FALSE;
		}
		if ( (*((XPFState *)state)->xpfMatchedNodes[((XPFState *)state)->xpfMatchedNum - 1])->nodetype & XPF_DAGS_HASHENTRY) {
			if (((XPFState *)state)->xpfMatchedNum < 2) return FALSE;	/* can't be! */
			(*((XPFState *)state)->xpfMatchedNodes[((XPFState *)state)->xpfMatchedNum - 2])->common.pattern.flags = flags & XPF_FLAG_USER;
			(*((XPFState *)state)->xpfMatchedNodes[((XPFState *)state)->xpfMatchedNum - 1])->common.hash.alarm = 
				(*((XPFState *)state)->xpfMatchedNodes[((XPFState *)state)->xpfMatchedNum - 2])->common.pattern.alarm = alarm;
		}
		else {
			if (idestfilter == NULL) {	/* can't be! */
				xpfRestoreHeaders(((XPFState *)state));
				return FALSE;
			}
			(*iproot)->common.pattern.flags = flags & XPF_FLAG_USER;
			(*iproot)->common.pattern.alarm = alarm;
		}

		xpfRestoreHeaders(((XPFState *)state));
		return TRUE;
	}
	return FALSE;
}

/*
 * Gets the flags and alarm at the matched pattern. There is no separate
 * function to get them independently, although one can write it if reqd.
 * Returns TRUE if it is able to get them properly, FALSE otherwise.
 */

int
xpfGetOptions(state, patternlist, pflags, palarm)
	XPFObject	state;
	XPFPattern	*patternlist;
	int		*pflags;
	short		*palarm;
{
	PATNode		**iproot = NULL;
	XPFFilter	**idestfilter = NULL;

	if (((XPFState *)state)->xpfAssociationSet != NULL) {
		/* Use this function to just get the set of matched nodes: no need to store headertype of patternlist since used only once */
		xpfGetPositionFromPatternList(state,
						patternlist,
						patternlist->filterlist,
						&((XPFState *)state)->xpfAssociationSet,
						&((XPFState *)state)->xpfAssociationSet->common.pattern.filterlist,
						((XPFState *)state)->xpfGranularity,
						XPF_DAGS_SRCNEWDESTNEW,
						&iproot,
						&idestfilter);

		if (((XPFState *)state)->xpfMatchedNum <= 0) {
			xpfRestoreHeaders(((XPFState *)state));
			return FALSE;
		}
		if ((*((XPFState *)state)->xpfMatchedNodes[((XPFState *)state)->xpfMatchedNum - 1])->nodetype & XPF_DAGS_HASHENTRY) {
			if (((XPFState *)state)->xpfMatchedNum < 2) return FALSE;	/* can't be! */
			*pflags = (*((XPFState *)state)->xpfMatchedNodes[((XPFState *)state)->xpfMatchedNum - 2])->common.pattern.flags;
			ERRPRINT(xpfAssert( (*((XPFState *)state)->xpfMatchedNodes[((XPFState *)state)->xpfMatchedNum - 1])->common.hash.alarm ==
				(*((XPFState *)state)->xpfMatchedNodes[((XPFState *)state)->xpfMatchedNum - 2])->common.pattern.alarm ));
			*palarm = (*((XPFState *)state)->xpfMatchedNodes[((XPFState *)state)->xpfMatchedNum - 1])->common.hash.alarm;
		}
		else {
			if (idestfilter == NULL) {	/* can't be! */
				xpfRestoreHeaders(((XPFState *)state));
				return FALSE;
			}
			*pflags = (*iproot)->common.pattern.flags;
			*palarm = (*iproot)->common.pattern.alarm;
		}

		xpfRestoreHeaders(((XPFState *)state));
		return TRUE;
	}
	return FALSE;
}

/*
 * Returns TRUE on success, FALSE on failure.
 */

int
xpfSetParameters(state, opcode, value)
	XPFObject	state;
	int		opcode;
	unsigned long	value;
{
	switch(opcode)
	{
	case XPF_SET_GETMSGCONTENTS:
		((XPFState *)state)->xpfGetMessageContents = value;
		return TRUE;
	case XPF_SET_PUTMSGCONTENTS:
		((XPFState *)state)->xpfPutMessageContents = value;
		return TRUE;
        case XPF_SET_GRANULARITY:
		if (value > XPF_DATA_MAXSIZE || value < sizeof(XPFDataUnit))
			return FALSE;
		((XPFState *)state)->xpfGranularity = value;
		return TRUE;
        case XPF_SET_MAXASSOCIATIONS:
		if (value <= 0) return FALSE;
		((XPFState *)state)->xpfMaxAssociations = value;
		return TRUE;
	default:
		return FALSE;
	}
}

/*
 * Returns TRUE on success, FALSE on failure.
 */

int
xpfGetParameters(state, opcode, pvalue)
	XPFState	*state;
	int		opcode;
	unsigned long	*pvalue;
{
	switch(opcode)
	{
	case XPF_GET_GETMSGCONTENTS:
		*pvalue = (unsigned long)((XPFState *)state)->xpfGetMessageContents;
		return TRUE;
	case XPF_GET_PUTMSGCONTENTS:
		*pvalue = (unsigned long)((XPFState *)state)->xpfPutMessageContents;
		return TRUE;
        case XPF_GET_MAXASSOCIATIONS:
		*pvalue = (int)((XPFState *)state)->xpfMaxAssociations;
		return TRUE;
        case XPF_GET_GRANULARITY:
		*pvalue = (int)((XPFState *)state)->xpfGranularity;
		return TRUE;
        case XPF_GET_NUMASSOCIATIONS:
		*pvalue = (int)((XPFState *)state)->xpfNumAssociations;
		return TRUE;
	default:
		return FALSE;
	}
}

