/*
 * xpf_interface.h: the only file the user needs to #include to use the
 *		interface. To specify the proper parameters the user has
 *		to look at xpf_user.h. xpf_imports.h contains 3 imported
 *		declarations used by xpf.
 */

#ifndef	_XPF_H_
#define _XPF_H_

#include "xpf_imports.h"
#include "xpf_user.h"

/*
 * This file describes the object oriented interface provided by XPF.
 * The code is reentrant. A sample usage of each interface function
 * is given in a comment above its declaration. A sample list of
 * variable declarations (used in functions below) is given here:
 *
 * XPFObject		x;
 * XPFInfoList		*deletedinfo;	-- ONLY if MERGEINFO
 * XPFMessageList	*deletedmsgs;	-- ONLY if DELAYEDMSGS && STOREMSGS
 * XPFMessageList	*delayedmsgs;	-- ONLY if DELAYEDMSGS && STOREMSGS
 * XPFInfoList		*associatedinfo;-- if !MERGEDINFO, use XPFInformation *associatedinfo;
 * XPFPattern		*p, *prefixpat, *newpat, *findsuffixpat;
 * XPFMessage		msg;
 * int			len, value, flags, alarm;
 * int			*pflags, *palarm;
 * int			ret;
 *
 * Two interface routines cannot be called simultaneously on the same object
 * "x" -- however, two routines CAN be called simultaneously on two different
 * xpf-objects x1 and x2. The user must synchronize her/his calls to xpf.
 * The behaviour of an interface function depends ONLY on its xpf-object x.
 *
 * The only functions xpf requires from "outside" are: xpfMalloc() and xpfFree().
 * They have been defined to be malloc and free in the file "xpf_imports.h"
 * In the protocols xpf_protl.c and xpftest.c, they are the standard x-Kernel routines.
 *
 * NOTE: The delayedmsgs list and deletedinfo list are used only when the
 * compilation flags STOREMSGS and MERGEINFO are used. Otherwise, they are not
 * required. For an explanation of the compiler flags, please see README.
 */

/*
 * SEMANTICS:
 * Must be called first. Returns the "state" which should be passed as
 * a parameter to all other interface calls. Two operations on one state
 * cannot be done in parallel (implicit serialization). However,
 * operations on two XPFs can be done in parallel: hence the device
 * driver can use one XPFObject for each network-interface.
 * Returns TRUE it xpf was able to allocate memory and intialize its
 * state, FALSE otherwise.
 *
 * USAGE:
 * if (xpfCreateObject(&x) == FALSE) exit(1);
 */
#if	PROTOTYPE
int	xpfCreateObject(XPFObject *px);
#else	PROTOTYPE
int	xpfCreateObject();
#endif	PROTOTYPE

/*
 * SEMANTICS:
 * Completely ruins the "state". The parameter is invalid after return.
 * Returns the set of deletedmessages and deletedinfos in specified pointers.
 * Has no return value.
 *
 * USAGE:
 * xpfDestroyObject(x, &deletedinfo, &deletedmsgs);
 *
 * for (; deletedinfo!=XPF_INFO_NULL;
 *	xDestroy(deletedinfo->information), -- if info is an xKernel session
 *      temp = deletedinfo,
 *      deletedinfo=deletedinfo->next,
 *	xpfFree(temp, sizeof(XPFInfoList)));
 *
 * for (; deletedmsgs!=NULL;
 *	msgDestroy(deletedmsgs->message),   -- if message is an xKernel message
 *	temp = deletedmsgs,
 *	deletedmsgs=deletedmsgs->next,
 *	xpfFree(temp, sizeof(XPFMessageList)));
 */
#if	PROTOTYPE
void	xpfDestroyObject(XPFObject x
#if	MERGEINFO
			,XPFInfoList **pdeletedinfo
#endif	MERGEINFO
#if	STOREMSGS && DELAYEDMSGS
			,XPFMessageList **pdeletedmsgs
#endif	STOREMSGS && DELAYEDMSGS
			);
#else	PROTOTYPE
void	xpfDestroyObject();
#endif	PROTOTYPE

/*
 * SEMANTICS:
 * Sets up an association between patternlist and info so that
 * messages matching the patternlist can be mapped to the info.
 * Returns TRUE if successful, FALSE otherwise.
 *
 * NOTE: This function might destroy the CONTENTS of the pattern's
 * header or filterlist. However, it DOES NOT change the linkage
 * information or ANYTHING that the USER might have allocated.
 * Hence, after this function returns, the user should reinitialize
 * all the filters in the pattern and the header. This is done due
 * to efficiency considerations: otherwise, xpf has to make a copy
 * of the parameters before Setup and destroy the copy after Setup
 * is complete, in ADDITION to the copies it makes DURING Setup.
 *
 * USAGE:
 * if (!xpfSetupAssociation(x, p, associatedinfo))
 *	printf("could not setup association!\n");
 */
#if	PROTOTYPE
int	xpfSetupAssociation(XPFObject x,
			    XPFPattern *patternlist
#if	MERGEINFO
			    ,XPFInfoList *associatedinfo
#else	MERGEINFO
			    ,XPFInformation associatedinfo
#endif	MERGEINFO
			    );
#else	PROTOTYPE
int	xpfSetupAssociation();
#endif	PROTOTYPE

/*
 * SEMANTICS:
 * Sets up an association between prefixpat.newpat and info
 * such that its suffix is the same as that of prefixpat.findsuffixpat.
 * This allows the user to manipulate patterns so that a logical OR
 * can be realized. Returns TRUE if successful, FALSE otherwise.
 *
 * NOTE-1: the headers in newpat and findsuffixpat must be identical.
 * Moreover, one or more of these patterns can be NULL subject to the
 * condition that findsuffixpat is NOT a prefix of newpat (cycle...).
 *
 * NOTE-2: the above facts about the CONTENTS of patterns are valid here too.
 *
 * USAGE:
 * if (!xpfSetupAssociationSuffix(x, prefix, new, findsuffix, associatedinfo))
 *	printf("error...\n");
 */
#if	PROTOTYPE
int	xpfSetupAssociationSuffix(XPFObject *state,
				  XPFPattern *prefixpat,
				  XPFPattern *newpat,
				  XPFPattern *findsuffixpat,
#if	MERGEINFO
				  XPFInfoList *associatedinfo
#else	MERGEINFO
				  XPFInformation associatedinfo
#endif	MERGEINFO
				  );
#else	PROTOTYPE
int	xpfSetupAssociationSuffix();
#endif	PROTOTYPE

/*
 * SEMANTICS:
 * Removes the association corr. to patternlist, sets *pdeletedinfo to
 * the info(s) that are no-longer associated with any pattern and
 * sets *pdeletedmsgs to the corr. list of delayed messages. Also
 * terminates programs (if any) generated by this pattern.
 *
 * Returns TRUE if the whole patternlist got deleted successfully;
 * FALSE o/wise -- pdeletedinfo and pdeletedmsgs might contain
 * deleted stuff EVEN if Remove was not entirely successful -- e.g., this
 * can happen if only a part of the patternlist->filterlist matched.
 *
 * NOTE: If the patternlist is "A and B", and "B" gets deleted, then
 * the info and messagelist associated with this pattern are
 * NOT returned. They are returned only when "A" also gets deleted
 * and the info and messagelist aren't associated w/ anything.
 * It is upto the user to figure out what patterns these correspond to
 * and destroy (free) them, re-establish new associations (setup),
 * or re-process the set of messages (lookup) through the filter.
 *
 * USAGE:
 * if (!xpfRemoveAssociation(x, p, &deletedinfo, &deletedmsgs))
 *	printf("could not remove association!\n");
 *
 * for (; deletedinfo!=XPF_INFO_NULL;
 *	xDestroy(deletedinfo->information), -- if info is an xKernel session
 *      temp = deletedinfo,
 *      deletedinfo=deletedinfo->next,
 *	xpfFree(temp, sizeof(XPFInfoList)));
 *
 * for (; deletedmsgs!=NULL;
 *	msgDestroy(deletedmsgs->message),   -- if message is an xKernel message
 *	temp = deletedmsgs,
 *	deletedmsgs=deletedmsgs->next,
 *	xpfFree(temp, sizeof(XPFMessageList)));
 */
#if	PROTOTYPE
int	xpfRemoveAssociation(XPFObject x,
			     XPFPattern *patternlist
#if	MERGEINFO
			     ,XPFInfoList **pdeletedinfo
#endif	MERGEINFO
#if	STOREMSGS && DELAYEDMSGS
			     ,XPFMessageList **pdeletedmsgs
#endif	STOREMSGS && DELAYEDMSGS
			     );
#else	PROTOTYPE
int	xpfRemoveAssociation();
#endif	PROTOTYPE

#if	DELAYEDMSGS && STOREMSGS
/*
 * SEMANTICS:
 * Removes the set of delayed messages at ALL patternlists that are
 * prefixes of the given "patternlist" and returns them in pdelayedmsgs.
 * These are EXACTLY those messages that could have been delayed due to
 * patterns in the given patternlist. The association between the patternlist
 * and its info still exists. Sets pdelayedmsgs and returns TRUE on
 * success; Returns FALSE with pdelayedmsgs untouched when no prefix
 * of patternlist has an association (failure).
 *
 * NOTE: The delayed messages returned in this case are EXACTLY those
 * that would have been returned by xpfRemoveAssociation called on every
 * prefix of the patternlist, the only difference being that the corr.
 * associations are NOT removed.
 *
 * USAGE:
 * if (xpfRemoveDelayedMessages(x, p, &delayedmsgs)) {
 *	-- EITHER call lookup on these messages, OR
 *	-- destroy these messages
 *	-- remember: the association corr. to "p" still exists
 * }
 */
#if	PROTOTYPE
int	xpfRemoveDelayedMessages(XPFObject x,
				 XPFPattern *patternlist,
				 XPFMessageList **pdelayedmsgs);
#else	PROTOTYPE
int	xpfRemoveDelayedMessages();
#endif	PROTOTYPE
#endif	DELAYEDMSGS && STOREMSGS

/*
 * SEMANTICS:
 * Returns XPF_USER_FAILURE, XPF_USER_DELAYED, or a combination of:
 * XPF_USER_MATCHEDNEW | XPF_USER_LAUNCHED | XPF_USER_TERMINATED.
 *
 * The pointer passociatedinfo is set to the information list associated with
 * the longest pattern that the message matches. However, if the message
 * is "delayed" for future processing (i.e., the return value is
 * XPF_USER_DELAYED), OR, there is no associated information for the
 * message or there is a serious system error (i.e., the return value is
 * XPF_USER_FAILURE), *passociatedinfo is unchanged.
 *
 * Whenever a process is triggered, it gets a higher priority than the
 * original set of patterns specified by the user for all future lookups
 * (i.e., existing processes are matched against before new ones are created).
 * Moreover, whenever a pattern triggers a process, the info associated
 * with that pattern is COPIED and THEN associated with the process (and
 * returned in *passociatedinfo) this gives the user a facility to CHANGE the
 * information (s) in the info associated with a process without effecting
 * the information mapped to the (parent) pattern that generated it.
 *
 * If the message matches a pattern that has no associated programs then
 * return value & XPF_USER_MATCHEDNEW will be == 0. If it matches a process
 * generated by the pattern, the above & will be 1. It is upto the user to
 * figure out whether or not this message was the first one that matched that
 * process by looking at the associated information returned by lookup (if it
 * is the 1st time, it'll be same as the info mapped to the generating pattern).
 *
 * The information mapped to previous processes if any is returned in the
 * pointer pdeletedinfo: if new processes were NOT generated, this
 * pointer is guaranteed to be NULL. Otherwise, it depends on the alarm values
 * specified by the user and whether any prcocesses were actually terminated.
 *
 * *pdelayedmsgs is set to a list of delayed messages encountered
 * while trying to match the message against patterns (the messages which can
 * "hope" * to match the newly generated processes). If no such messages are
 * encountered, *pdelayedmsgs is NULL. If the message does not spawn
 * processes (i.e., return value & XPF_USER_LAUNCHED == 0), *pdelayedmsgs
 * is always NULL; if it spawns a fresh set of processes (the above & == 1)
 * then *pdelayedmsgs may or may not be NULL.
 *
 * USAGE:
 * function xpftest_lookup(msg, len) {
 *	-- set locks
 *	ret = xpfLookupAssociation(x, msg, len, &associatedinfo, &deletedinfo, &delayedmsgs);
 *
 *	if ((ret & XPF_USER_DELAYED) || (ret & XPF_USER_FAILURE)) {
 *		-- release locks
 *		if (ret & XPF_USER_FAILURE) return XK_FAILURE;
 *		return XK_SUCCESS;
 *	}
 *
 * 	for (; deletedinfo!=XPF_INFO_NULL;
 *	    xDestroy(deletedinfo->information),
 *	    temp = deletedinfo,
 *	    deletedinfo=deletedinfo->next,
 *	    xpfFree(temp, sizeof(XPFInfoList)));
 *	-- this is the info associated with destroyed processes (if any)
 *	   which depends on the alarm-period specified in the pattern
 *	   that generated the process. REMEMBER: the user CANNOT setup
 *	   or remove processes directly: they are indirectly setup thru
 *	   programs and are removed via alarms (specified in patterns).
 *
 *	for(;associatedinfo!=XPF_INFO_NULL; associatedinfo=associatedinfo->next)
 *	    xpftest_process_info(associatedinfo->information);
 *
 *	while(delayedmsgs!=NULL)
 *	    xpftest_lookup(delayedmsgs->message, delayedmsgs->length);
 *	-- do recursive calls; you can also queue them up for later processing
 *
 *	-- release locks
 *	return (XK_SUCCESS);
 * }
 */
#if	PROTOTYPE
int	xpfLookupAssociation(XPFObject x,
			     XPFMessage message,
			     int length
#if	MERGEINFO
			     ,XPFInfoList **passociatedinfo
#else	MERGEINFO
			     ,XPFInformation *passociatedinfo
#endif	MERGEINFO
#if	MERGEINFO
			     ,XPFInfoList **pdeletedinfo
#endif	MERGEINFO
#if	STOREMSGS && DELAYEDMSGS
			     ,XPFMessageList **pdelayedmsgs
#endif	STOREMSGS && DELAYEDMSGS
			     );
#else	PROTOTYPE
int	xpfLookupAssociation();
#endif	PROTOTYPE

/*
 * SEMANTICS:
 * Sets the flags and alarm at the matched pattern. There is no separate
 * function to set them independently, although one can write it if reqd.
 * Returns TRUE if it is able to set them properly, FALSE otherwise.
 *
 * USAGE:
 * if (!xpfSetOptions(x, patternlist, XPF_FLAG_NOSPECIFIC|XPF_FLAG_DELAY, >=0))
 *	printf("error...\n");
 * -- an alarm value of 0 => creation of processes is disabled.
 *    an alarm value of N >=1 => When the N+1th process is created, the Nth
 *    one is destroyed (maintenance is in FIFO order).
 *    an alarm can therefore be interpreted as the number of concurrent
 *    processes that should exist at one time.
 * -- a flag value of XPF_FLAG_DELAY at a pattern => if a message DOES-NOT
 *    match that pattern, instead of trying to look for the longest-matching
 *    pattern, xpf queues up that message at that pattern until a message that
 *    matches that pattern actually arrives. A flag value of XPF_FLAG_NOSPECIFIC
 *    disables the setting-up of new patterns whose filterlist is MORE SPECIFIC
 *    than the filter-list of the given patternlist.
 *
 * DEFAULTS: flag = 0, alarm = 1.
 */
#if	PROTOTYPE
int	xpfSetOptions(XPFObject x,
		      XPFPattern *patternlist,
		      int flags,
		      short alarm);
#else	PROTOTYPE
int	xpfSetOptions();
#endif	PROTOTYPE

/*
 * SEMANTICS:
 * Gets the flags and alarm at the matched pattern. There is no separate
 * function to get them independently, although one can write it if reqd.
 * Returns TRUE if it is able to get them properly, FALSE otherwise.
 *
 * USAGE:
 * if (!xpfGetOptions(x, patternlist, pflags, palarm)) printf("error...\n");
 */
#if	PROTOTYPE
int	xpfGetOptions(XPFObject x,
		      XPFPattern *patternlist,
		      int *pflags,
		      short *palarm);
#else	PROTOTYPE
int	xpfGetOptions();
#endif	PROTOTYPE

/*
 * SEMANTICS:
 * Returns TRUE on success, FALSE on failure.
 *
 * USAGE:
 * if (!xpfSetParameters(x,
 *			XPF_SET_GRANULARITY or
 *			XPF_SET_MAXASSOCIATIONS or
 *			XPF_SET_GETMSGCONTENTS or
 *			XPF_SET_PUTMSGCONTENTS,
 *			value)) printf("error ...\n");
 *
 * DEFAULTS: max-associations = 1024, granularity = 2 (see xpf_dep.h).
 */
#if	PROTOTYPE
int	xpfSetParameters(XPFObject x, int opcode, int value);
#else	PROTOTYPE
int	xpfSetParameters();
#endif	PROTOTYPE

/*
 * SEMANTICS:
 * Returns TRUE on success, FALSE on failure.
 *
 * USAGE:
 * if (!xpfGetParameters(x,
 *			XPF_GET_GRANULARITY or
 *			XPF_GET_MAXASSOCIATIONS or
 *			XPF_GET_NUMASSOCIATIONS or
 *			XPF_GET_GETMSGCONTENTS or
 *			XPF_GET_PUTMSGCONTENTS,
 *			&value)) printf("error ...\n");
 * else -- look at the value and perhaps do an xpfSetParameters;
 */
#if	PROTOTYPE
int	xpfGetParameters(XPFObject x, int opcode, int *pvalue);
#else	PROTOTYPE
int	xpfGetParameters();
#endif	PROTOTYPE

/*
 * Sample declarations for the functions to get/put contents of a message.
 * ------------------------------------------------------------------------
 * These are the only "up-calls" done by xpf (apart from xpfMalloc and xpfFree).
 *
 * If they are NULL (SetParameters(x, XPF_SET_GETMSGCONTENTS, getfuncptr) is
 * NOT done after creating and xpf-object x), then the message is assumed to
 * be a char* pointer and data is accessed from the message assuming it is an
 * array of characters. In this simple case, the user doesn't have to specify
 * any put-contents function at all (the user doesn't have to do a
 * SetParameters(x, XPF_SET_PUTMSGCONTENTS, putfuncptr)).
 *
 * These however make sense in the context of the xKernel where a message is
 * complex structure (stack + tree...) where we need special functions to
 * get/put data. Sample declarations of getfuncptr and putfuncptr are:
 *
 *	XPFDataUnit	*(*getfuncptr)(XPFMessage message,
 *					int message_length,
 *					int size_extracted,
 *					int size_needed);
 *
 * - the parameters tell the user: the original size of the message (as passed
 *   to xpfLookupAssociation), the amount already extracted from it and the
 *   requested size (as headers suggest). If this function returns a "NULL"
 *   buffer, then xpf interprets this as an error (and returns XPF_USER_FAILURE
 *   or FALSE from the parent interface function).
 *
 *	void	(*putfuncptr)(XPFMessage message, 
 *				int message_length,
 *				int size_extracted,
 *				int size_notneeded,
 *				XPFDataUnit *buffer);
 *
 * - the parameters tell the user: the original size of the message, the size
 *   extracted so far and the amount of data being given back (i.e., size of
 *   data that is not needed anymore, which is the size of the data pointed
 *   to by "buffer").
 *
 * - xpf maintains the current-amount-extracted in its state so that it can
 *   avoid copying when these functions are NULL: the user may/may not use it.
 */

#endif	_XPF_H_
